/*
# $BEGIN_KEYHOLE_LICENSE$
# 
# This file is part of the Keyhole project, a lightweight library for
# the Arduino IDE for interpreting commands and communicating variable
# values via a JSON- and Python-compatible text interface. 
# 
# Copyright (c) 2023- Jeremy Hill
# 
# Keyhole is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/ .
# 
# $END_KEYHOLE_LICENSE$
*/
#ifndef   __Keyhole_H__
#define   __Keyhole_H__

#include "Arduino.h" // for Stream and String
/*
  The following example allows two global variables to be queried and/or
  set over a serial port or other Stream::
  
      #include "Keyhole.h"
    
      double foo = 0.0;
      String bar = "look at these escaped characters: \" \x08 \\ \t\r\n";
    
      void setup()
      {
        Serial.begin();
      }
    
      void loop()
      {
        // main business of your sketch goes here
        // ...
        
        KEYHOLE keyhole(Serial);
        if( keyhole.begin() )
        {
          if(keyhole.variable("foo", foo)) Serial.println("assigned to foo");
          if(keyhole.variable("bar", bar)) Serial.println("assigned to bar");
          if(keyhole.command("Marco!")) Serial.println("Polo!");
          keyhole.end();
        }
      }
      
  An optional third argument to `variable()` can be `VARIABLE_READ_ONLY`,
  `VARIABLE_VERBOSE`, or the default value `VARIABLE_SILENT`.
  
  In this example, if you send the command `foo` over the serial port, the
  reply will be a JSON string associating the name `foo` with the current
  value of the corresponding global variable `gFoo`. 
  
  Provided the variable is not designated as `VARIABLE_READ_ONLY`, you can
  also set its value, e.g. by sending the command `foo = 2`.  (If you have
  designated the variable as `VARIABLE_VERBOSE` then this will also trigger
  a JSON output, as if you had queried `foo` immediately after setting it.)

  You can send the simple command `?` to receive a JSON output containing all
  the variables that are accessible in this manner.

  You can also have the full report delivered automatically on a repeating
  schedule by setting the `.autoSeconds` member greater than zero; you can
  even allow this parameter itself to be controlled via the keyhole, by
  exposing the `autoSeconds` member itself as a variable::

        KEYHOLE keyhole(Serial);
        if(keyhole.begin())
        {
          if(keyhole.variable("foo", foo)) Serial.println("assigned to foo");
          if(keyhole.variable("bar", bar)) Serial.println("assigned to bar");
          keyhole.variable("report_period_sec", keyhole.autoSeconds);
          keyhole.end();
        }

  Under the hood: `keyhole` is an instance of the class `Keyhole`. Such
  instances must either be global variables, or declared `static`. The
  `KEYHOLE` macro can be used inside or outside of `loop()` - it simply
  declares a `static Keyhole` instance (while hiding the possibly-not-so-
  user-friendly keyword `static` from users who are not familiar with it).
  
*/

// Debugging macros:
#define DBSTREAM Serial
#define REPORT( X )  { DBSTREAM.print( "{\"" #X "\" : " ); DBSTREAM.print( X );                         DBSTREAM.println( "}" ); DBSTREAM.flush(); }
#define REPORTS( X ) { DBSTREAM.print( "{\"" #X "\" : " ); Keyhole( DBSTREAM ).printStringLiteral( X ); DBSTREAM.println( "}" ); DBSTREAM.flush(); }

typedef enum
{
	VARIABLE_READ_ONLY = 0,
	VARIABLE_SILENT    = 1,
	VARIABLE_VERBOSE   = 2
} VariableWriteMode;

#define KEYHOLE       static Keyhole
class Keyhole
{
	public:
		Keyhole( Stream & stream=Serial, float autoSeconds=0.0 );
		~Keyhole();
	
		bool begin( void );
		bool begin( unsigned long microsecondTimestamp );
	
		bool command( const char * cmd );
	
		bool variable( const char * variableName, int &            referenceToVariable,  VariableWriteMode mode=VARIABLE_SILENT );
		bool variable( const char * variableName, unsigned int &   referenceToVariable,  VariableWriteMode mode=VARIABLE_SILENT );
		bool variable( const char * variableName, short &          referenceToVariable,  VariableWriteMode mode=VARIABLE_SILENT );
		bool variable( const char * variableName, unsigned short & referenceToVariable,  VariableWriteMode mode=VARIABLE_SILENT );
		bool variable( const char * variableName, long &           referenceToVariable,  VariableWriteMode mode=VARIABLE_SILENT );
		bool variable( const char * variableName, unsigned long &  referenceToVariable,  VariableWriteMode mode=VARIABLE_SILENT );
		bool variable( const char * variableName, float &          referenceToVariable,  VariableWriteMode mode=VARIABLE_SILENT );
		bool variable( const char * variableName, double &         referenceToVariable,  VariableWriteMode mode=VARIABLE_SILENT );
		bool variable( const char * variableName, String &         referenceToVariable,  VariableWriteMode mode=VARIABLE_SILENT );
#		define variableAssigned variable
	
		bool end( void );
	
		unsigned long elapsed( void );
		void          printStringLiteral( const String & s, bool withQuotes=true );

		Stream &      stream;
		float         autoSeconds;
		
	private:
		unsigned long mBeginMicros;
		String        mFullCommand;
		int           mListAllVariables;
		String        mPartialCommand;
		bool          mBackslash;
		int           mHexEscape;
		char          mHexValue;
		char          mQuote;
		unsigned long mTimestampOfLastAutoReport;

		const char *  _parseVariableCommand( const char * variableName, int & commandLength );
};

unsigned long StrToUnsignedInteger( const char * start, char ** endptr );
long          StrToSignedInteger(   const char * start, char ** endptr );
double        StrToDouble( const char * start, char ** endptr );

#endif // __Keyhole_H__
